#include "a4c.ime.h"
#include "a4c.tk.h"

void SwapEndian(char *v)
{
  char t;
  #define SW(n,w) t=*(n); *(n)=*(w); *(w)=t;
  SW(v,v+3); SW(v+1,v+2);
  #undef  SW
}

void nope(char * fmt,...)
{
  /* do nothing */
}

void oops(char * fmt,...)
{
  va_list argptr;
  va_start(argptr,fmt);
  vfprintf(stderr,fmt,argptr);
  va_end(argptr);
  fprintf(stderr,"\n");
}

void die(char * fmt,...)
{
  va_list argptr;
  va_start(argptr,fmt);
  vfprintf(stderr,fmt,argptr);
  va_end(argptr);
  fprintf(stderr,"\n");
  exit(-1);
}

char * xDisplay(void)
{
  char *display;
  display = getenv("DISPLAY");
  if (display==NULL) display = getenv("display");
  return display;
}

int xChangeTitle(Display * disp, Window win, char *title)
{
  XTextProperty text;
  if (XStringListToTextProperty(&title,1,&text)!=0) {
    XSetWMName(disp,win,&text);
    XFree(text.value);
    return 1;
  } else {
    oops("a4c: Can not change title to %s",title);
    return 0;
  }
}

int a4cPlayWAVE(char *DEVICE,char *FILENAME)
{
  FILE * WAV;
  FILE * AUDIO;
  int    c;

  WAV   = fopen(FILENAME,"rb");
  if (WAV==NULL) {
    report("Could not open %s file.",FILENAME);
    return 0;
  }
  AUDIO = fopen(DEVICE,"wb");
  if (AUDIO==NULL) {
    report("Could not open %s device.",DEVICE);
    return 0;
  }
  while ((c=fgetc(WAV))!=EOF) fputc(c,AUDIO);
  fclose(WAV);
  fclose(AUDIO);
  return 1;
}

unsigned long a4cXCOLOR(a4cMAIN * MAIN,char *name)
{
  XColor color;
  if (!XParseColor(MAIN->disp,MAIN->COLORMAP,name,&color))
    die("I don't think we have this color : %s",name);
  if (!XAllocColor(MAIN->disp,MAIN->COLORMAP,&color)) 
    die("Oh! It's too bad, %s is out of stock.",name);
  return color.pixel;
}

void a4cAllocXCOLOR(a4cMAIN *MAIN)
{
  #define LOADCOLOR(c,n) MAIN->ENVCOLOR.c = a4cXCOLOR(MAIN,MAIN->ENVCOLOR.n)
  LOADCOLOR(foreground,fgcolor);
  LOADCOLOR(background,bgcolor);
  LOADCOLOR(lightcolor,ltcolor);
  LOADCOLOR(darkcolor ,dkcolor);
  LOADCOLOR(englishcolor,egcolor);
  #undef  LOADCOLOR
}

int isInRECT(a4cRECT *RECT,int x,int y)
{
  return (RECT->x1<=x&&
          RECT->x2>=x&&
          RECT->y1<=y&&
          RECT->y2>=y);
}

int isInKeyBox(a4cKeyBox *KEYBOX,int x,int y)
{
  return isInRECT(&KEYBOX->Border,x,y);
}

void a4cErrorHandler(Display *disp, XErrorEvent *event)
{
  die("Encounter a X Error Event!");
}

void a4cInitGC(a4cMAIN *MAIN)
{
  XGCValues gcv;
  int flags = GCForeground | GCBackground | GCFont | GCGraphicsExposures;
  gcv.foreground = MAIN->ENVCOLOR.foreground;
  gcv.background = MAIN->ENVCOLOR.background;
  gcv.graphics_exposures = 0;
  gcv.font  = MAIN->EFONT->fid;
  MAIN->EGC = XCreateGC(MAIN->disp,MAIN->win,flags,&gcv);
  if (MAIN->EGC==NULL) die("Can not create EGC!");
  gcv.font  = MAIN->CFONT->fid;
  MAIN->CGC = XCreateGC(MAIN->disp,MAIN->win,flags,&gcv);
  if (MAIN->CGC==NULL) die("Can not create CGC!");
}

void a4cLoadFonts(a4cMAIN *MAIN,char *efont,char *cfont)
{
  MAIN->EFONT = XLoadQueryFont(MAIN->disp,efont);
  if (!MAIN->EFONT) die("Can not load font %s",efont);
  MAIN->CFONT = XLoadQueryFont(MAIN->disp,cfont);
  if (!MAIN->CFONT) die("Can not load font %s",cfont);
}

void a4cInitRanges(a4cMAIN * PAM)
{
  int cwidth,cheight;
  int ewidth,eheight;
  int width,height;
  
  cwidth  = PAM->CFONT->max_bounds.width;
  cheight = PAM->CFONT->ascent + PAM->CFONT->descent;

  ewidth  = PAM->EFONT->max_bounds.width *2;
  eheight = PAM->EFONT->ascent + PAM->EFONT->descent;

  width   = (cwidth >ewidth ) ? cwidth :ewidth;
  height  = (cheight>eheight) ? cheight:eheight;

  /**********************************************************\
               All For Chinese Interface Arrangement
   +-------------------------------------------------------+
   |+-----------------------------------------------------+|
   ||                       ListArea                      ||
   |+-----------------------------------------------------+|
   |+------++--------------------------++-------++--------+|
   ||METHOD||          KeyArea         ||CHARSET||FONTSIZE||
   |+------++--------------------------++-------++--------+|
   +-------------------------------------------------------+

   A4C_METHOD_WIDTH   5
   A4C_CHARSET_WIDTH  2
   A4C_FONTSIZE_WIDTH 2
   A4C_KEYAREA_WIDTH  10
   A4C_MARGAIN        2

  \**********************************************************/
   PAM->Border.x1   = 0; 
   PAM->Border.y1   = 0;

   PAM->ListArea.Border.x1 = 4; 
   PAM->ListArea.Border.y1 = 4;
   PAM->ListArea.Text.x1   = PAM->ListArea.Border.x1 + 3;
   PAM->ListArea.Text.y1   = PAM->ListArea.Border.y1 + 3;
   PAM->ListArea.Text.y2   = PAM->ListArea.Text.y1 + height;
   PAM->ListArea.Border.y2 = PAM->ListArea.Text.y2 + 3;

   PAM->Method.Border.x1   = 4;
   PAM->Method.Border.y1   = PAM->ListArea.Border.y2 + 2;
   PAM->Method.Text.x1     = PAM->Method.Border.x1 + 3;
   PAM->Method.Text.y1     = PAM->Method.Border.y1 + 3;
   PAM->Method.Text.y2     = PAM->Method.Text.y1 + height;
   PAM->Method.Border.y2   = PAM->Method.Text.y2 + 3;
   PAM->Method.Text.x2     = PAM->Method.Text.x1 + A4C_METHOD_WIDTH * width;
   PAM->Method.Border.x2   = PAM->Method.Text.x2 + 3;

   PAM->Border.y2          = PAM->Method.Border.y2 + 4;

   PAM->KeyArea.Border.x1  = PAM->Method.Border.x2 + 2;
   PAM->KeyArea.Border.y1  = PAM->Method.Border.y1;
   PAM->KeyArea.Text.x1    = PAM->KeyArea.Border.x1 + 3;
   PAM->KeyArea.Text.y1    = PAM->Method.Text.y1;
   PAM->KeyArea.Text.y2    = PAM->Method.Text.y2;
   PAM->KeyArea.Border.y2  = PAM->Method.Border.y2;
   PAM->KeyArea.Text.x2    = PAM->KeyArea.Text.x1 + A4C_KEYAREA_WIDTH * width;
   PAM->KeyArea.Border.x2  = PAM->KeyArea.Text.x2 + 3;

   PAM->Charset.Border.x1  = PAM->KeyArea.Border.x2 + 2;
   PAM->Charset.Border.y1  = PAM->Method.Border.y1;
   PAM->Charset.Text.x1    = PAM->Charset.Border.x1 + 3;
   PAM->Charset.Text.y1    = PAM->Method.Text.y1;
   PAM->Charset.Text.y2    = PAM->Method.Text.y2;
   PAM->Charset.Border.y2  = PAM->Method.Border.y2;
   PAM->Charset.Text.x2    = PAM->Charset.Text.x1 + A4C_CHARSET_WIDTH * width;
   PAM->Charset.Border.x2  = PAM->Charset.Text.x2 + 3;

   PAM->FontSize.Border.x1 = PAM->Charset.Border.x2 + 2;
   PAM->FontSize.Border.y1 = PAM->Method.Border.y1;
   PAM->FontSize.Text.x1   = PAM->FontSize.Border.x1 + 3;
   PAM->FontSize.Text.y1   = PAM->Method.Text.y1;
   PAM->FontSize.Text.y2   = PAM->Method.Text.y2;
   PAM->FontSize.Border.y2 = PAM->Method.Border.y2;
   PAM->FontSize.Text.x2   = PAM->FontSize.Text.x1 + A4C_FONTSIZE_WIDTH * width;
   PAM->FontSize.Border.x2 = PAM->FontSize.Text.x2 + 3;

   PAM->ListArea.Text.x2   = PAM->FontSize.Text.x2;
   PAM->ListArea.Border.x2 = PAM->FontSize.Border.x2;

   PAM->Border.x2          = PAM->ListArea.Border.x2 + 4;

   PAM->ListArea.style = A4C_STYLE_NOBORDER;
   PAM->FontSize.style = A4C_STYLE_CENTER;
   PAM->KeyArea.style  = 0;
   PAM->Charset.style  = A4C_STYLE_CENTER;
   PAM->Method.style   = A4C_STYLE_CENTER;
}


void a4cLoadAccelerators(char *filename,a4cMAIN * MAIN)
{
  FILE * fp;
  iomeConfig CONF;
  int n;
  int state;
  int c;
  char * buf;
  char * bx;

  memset(&CONF,0,sizeof(CONF));
  iomeLoadConfiguration(filename,&CONF);
  
  MAIN->AccelKeys = 0;
  for (n=0;n<CONF.size;n++) {
    
    buf = CONF.items[n].key;
    state = 0;
    
    bx  = buf;
    while (*buf && *buf!='+') buf++;
    if (*buf=='+') {
      *buf = 0;
      buf++;
      if (strcasecmp(bx,"ctrl")==0)  state |= IME_KF_CONTROL;
      if (strcasecmp(bx,"shift")==0) state |= IME_KF_SHIFT;
      if (strcasecmp(bx,"alt")==0)   state |= IME_KF_ALT;
    } else continue;
    
    bx = buf;
    while (*buf && *buf!='+') buf++;
    if (*buf=='+') {
      *buf = 0;
      buf++;
      if (strcasecmp(bx,"ctrl")==0)  state |= IME_KF_CONTROL;
      if (strcasecmp(bx,"shift")==0) state |= IME_KF_SHIFT;
      if (strcasecmp(bx,"alt")==0)   state |= IME_KF_ALT;
    } else continue;

    c = *buf;
    
    buf = CONF.items[n].value;
    if (isprint(c)) {
      if (*buf=='"') {
        buf++;
        bx = buf;
        while (*buf && *buf!='"') buf++;
        *buf = 0;
        MAIN->ACCEL[MAIN->AccelKeys].state   = state;
        MAIN->ACCEL[MAIN->AccelKeys].key     = tolower(c);
        MAIN->ACCEL[MAIN->AccelKeys].method  = -1;
        MAIN->ACCEL[MAIN->AccelKeys].command = strdup(bx);
        MAIN->AccelKeys++;
      } else {
        bx = buf;
        MAIN->ACCEL[MAIN->AccelKeys].state   = state;
        MAIN->ACCEL[MAIN->AccelKeys].key     = tolower(c);
        MAIN->ACCEL[MAIN->AccelKeys].method  = 0;
        MAIN->ACCEL[MAIN->AccelKeys].command = strdup(bx);
        MAIN->AccelKeys++;
      }
    }
  }
  
  iomeFreeConfiguration(&CONF);
}

int a4cLookupAccelerator(int state,int ch,a4cMAIN * MAIN)
{
  int n;
  int c = tolower(ch);
  
  for (n=0;n<MAIN->AccelKeys;n++)
    if (MAIN->ACCEL[n].state==state&&MAIN->ACCEL[n].key==c) return n;
  
  return -1;
}

int a4cWarmUp(
               int argc,
               char **argv,
               a4cMAIN * MAIN,
               iomeDefaultConf *CONF,
               char **pixmap
             )
{
  XSizeHints DefaultSizeHint = {
    PMinSize | PMaxSize | PResizeInc | PBaseSize | PWinGravity,
    0, 0, 480, 48,
    0, 0,
    480, 48,
    480, 48,
    {0,0},{0,0},
    480,48,
    NorthWestGravity
  };
  char *        DISPNAME;
  char *        efont;
  char *        cfont;
  XTextProperty name;
  char *        myname;
  int           w;
  int           h;
  Window        twin;
  char *        lang;

  MAIN->Endian = IS_ENDIAN();
  DISPNAME = iomeWhichItem(&CONF->SYS,"display");
  if (DISPNAME==NULL) DISPNAME = xDisplay();
  if (DISPNAME==NULL) die("No display specified!");
  MAIN->disp = XOpenDisplay(DISPNAME);
  if (!MAIN->disp) die("Oops, can not open display name %s",DISPNAME);

  XSetErrorHandler((XErrorHandler)a4cErrorHandler);

  MAIN->screen   = DefaultScreen(MAIN->disp);
  MAIN->root     = RootWindow(MAIN->disp,MAIN->screen);
  MAIN->COLORMAP = DefaultColormap(MAIN->disp,MAIN->screen);

  MAIN->ENVCOLOR.fgcolor = iomeWhichItem(&CONF->CONF,"fgcolor");
  MAIN->ENVCOLOR.bgcolor = iomeWhichItem(&CONF->CONF,"bgcolor");
  MAIN->ENVCOLOR.ltcolor = iomeWhichItem(&CONF->CONF,"ltcolor");
  MAIN->ENVCOLOR.dkcolor = iomeWhichItem(&CONF->CONF,"dkcolor");
  MAIN->ENVCOLOR.egcolor = iomeWhichItem(&CONF->CONF,"egcolor");
  
  a4cAllocXCOLOR(MAIN);

  efont = iomeWhichItem(&CONF->CONF,"font.english");
  cfont = iomeWhichItem(&CONF->CONF,"font.chinese");

  a4cLoadFonts(MAIN,efont,cfont);

  a4cInitRanges(MAIN);

  w = MAIN->Border.x2 + 1;
  h = MAIN->Border.y2 + 1;

  MAIN->win = XCreateSimpleWindow(
                MAIN->disp,
                MAIN->root,
                0,0,w,h,0,
                MAIN->ENVCOLOR.foreground,
                MAIN->ENVCOLOR.background
              );


  XGetWindowAttributes(MAIN->disp,MAIN->win,&MAIN->attr);

  XpmCreatePixmapFromData(
    MAIN->disp,
    MAIN->win,
    pixmap,
    &MAIN->ICON,
    0,
    (XpmAttributes*)&MAIN->attr
  );

  myname = iomeWhichItem(&CONF->CONF,"title");
  MAIN->title = myname;
  MAIN->CLASS.res_name   = myname;
  MAIN->CLASS.res_class  = "iome";

  MAIN->wm.input         = False;
  MAIN->wm.initial_state = NormalState;
  MAIN->wm.flags         = InputHint | StateHint | IconPixmapHint;
  MAIN->wm.icon_pixmap   = MAIN->ICON;

  MAIN->SIZE                  = DefaultSizeHint;
  MAIN->SIZE.width            = w;
  MAIN->SIZE.height           = h;
  MAIN->SIZE.max_width        = w;
  MAIN->SIZE.max_height       = h;
  MAIN->SIZE.width_inc        = w;
  MAIN->SIZE.height_inc       = h;
  MAIN->SIZE.base_width       = w;
  MAIN->SIZE.base_height      = h;

  if (XStringListToTextProperty(&myname,1,&name)==0) die("No Window Name!");

  XSetWMProperties(
    MAIN->disp,
    MAIN->win,
    &name,
    &name,
    argv,
    argc,
    &MAIN->SIZE,
    &MAIN->wm,
    &MAIN->CLASS
  );

  XFree(name.value);

  XSelectInput(MAIN->disp,MAIN->win,A4C_EVENTS);
  XMapWindow(MAIN->disp,MAIN->win);

  MAIN->XCIN = XInternAtom(MAIN->disp,A4C_XIN_ATOM,False);
  twin = XGetSelectionOwner(MAIN->disp,MAIN->XCIN);
  if (twin!=None) die("There exists a copy of a4c.ime!");
  XSetSelectionOwner(MAIN->disp,MAIN->XCIN,MAIN->win,CurrentTime);

  a4cInitGC(MAIN);
  
  MAIN->Method.message   = iomeWhichItem(&CONF->CONF,"method.english");
  MAIN->Charset.message  = iomeWhichItem(&CONF->CONF,"charset.neutral");
  MAIN->FontSize.message = iomeWhichItem(&CONF->CONF,"size.half");
  MAIN->KeyArea.message  = (char*)malloc(1024);
  memset(MAIN->KeyArea.message,0,1024);
  MAIN->ListArea.message = (char*)malloc(1024);
  memset(MAIN->ListArea.message,0,1024);
  strcpy(MAIN->ListArea.message,MAIN->title);

  MAIN->WM_DEL_WINDOW = XInternAtom(MAIN->disp,"WM_DELETE_WINDOW",False);
  XSetWMProtocols(MAIN->disp,MAIN->win,&MAIN->WM_DEL_WINDOW,1);

  xChangeTitle(MAIN->disp,MAIN->win,MAIN->title);
  
  lang = iomeWhichItem(&CONF->SYS,"lang");
  if (strcmp(lang,"big5")==0) {
    MAIN->CC           = isBig5;
  } else
  if (strcmp(lang,"gb2312")==0) {
    MAIN->CC = isGB2312;
  } else
    die("Well, bogie!");
  MAIN->base_charset = lang;

  return 1;
}

void a4cXWAIT(a4cEventHandler *a4c,a4cALL *ALL)
{
  XEvent event;

  #define CASTING(item,evt)\
     case item: a4c->evt(event,ALL); break

  while (1) {
    XNextEvent(ALL->MAIN.disp,&event);
    switch (event.type) {
      CASTING(KeyPress,keypress);
      CASTING(ClientMessage,clientmessage);
      CASTING(PropertyNotify,propertynotify);
      CASTING(MappingNotify,mappingnotify);
      CASTING(GraphicsExpose,graphicsexpose);
      CASTING(Expose,expose);
      CASTING(ButtonPress,buttonpress);
      CASTING(VisibilityNotify,visibilitynotify);
      CASTING(DestroyNotify,destroynotify);
      default: a4c->unknown(event,ALL);
    }
  }
  #undef  CASTING
}

void a4cKeyPress(XEvent event,a4cALL *ALL)
{
  if (!ALL->MAIN.Hide) XRaiseWindow(ALL->MAIN.disp,ALL->MAIN.win);
}

void a4cMappingNotify(XEvent event,a4cALL *ALL)
{
  XRefreshKeyboardMapping(&event.xmapping);
}

void a4cUnknownEvent(XEvent event,a4cALL *ALL)
{
  /* do nothing */
  /* report("How it that, %d?",event.type); */
}


void a4cDrawShadowBox(a4cMAIN *MAIN,a4cRECT * RECT, int up_down)
{
  XSetForeground(
    MAIN->disp,
    MAIN->CGC,
    (up_down)?MAIN->ENVCOLOR.lightcolor:MAIN->ENVCOLOR.darkcolor
  );
  XDrawLine(MAIN->disp,MAIN->win,MAIN->CGC,RECT->x1  ,RECT->y1  ,RECT->x2  ,RECT->y1  );
  XDrawLine(MAIN->disp,MAIN->win,MAIN->CGC,RECT->x1  ,RECT->y1+1,RECT->x2  ,RECT->y1+1);
  XDrawLine(MAIN->disp,MAIN->win,MAIN->CGC,RECT->x1  ,RECT->y1  ,RECT->x1  ,RECT->y2  );
  XDrawLine(MAIN->disp,MAIN->win,MAIN->CGC,RECT->x1+1,RECT->y1  ,RECT->x1+1,RECT->y2  );

  XSetForeground(
    MAIN->disp,
    MAIN->CGC,
    (up_down)?MAIN->ENVCOLOR.darkcolor:MAIN->ENVCOLOR.lightcolor
  );
  XDrawLine(MAIN->disp,MAIN->win,MAIN->CGC,RECT->x2  ,RECT->y1  ,RECT->x2  ,RECT->y2  );
  XDrawLine(MAIN->disp,MAIN->win,MAIN->CGC,RECT->x2-1,RECT->y1+1,RECT->x2-1,RECT->y2  );
  XDrawLine(MAIN->disp,MAIN->win,MAIN->CGC,RECT->x1  ,RECT->y2  ,RECT->x2  ,RECT->y2  );
  XDrawLine(MAIN->disp,MAIN->win,MAIN->CGC,RECT->x1+1,RECT->y2-1,RECT->x2  ,RECT->y2-1);
}

void a4cDrawTextInBox(a4cMAIN *MAIN,a4cKeyBox *KeyBox)
{
  int x,y,w;
  int ewidth,cwidth;
  int eAscent,cAscent;
  char * p = KeyBox->message;
  int EnableChinese;
  int isGB;
  char GB[3];

  XSetForeground(MAIN->disp,MAIN->EGC,MAIN->ENVCOLOR.englishcolor);
  XSetBackground(MAIN->disp,MAIN->EGC,MAIN->ENVCOLOR.background);
  XSetForeground(MAIN->disp,MAIN->CGC,MAIN->ENVCOLOR.foreground);
  XSetBackground(MAIN->disp,MAIN->CGC,MAIN->ENVCOLOR.background);

  x = KeyBox->Text.x1;
  y = KeyBox->Text.y1;

  ewidth = MAIN->EFONT->max_bounds.width;
  cwidth = MAIN->CFONT->max_bounds.width;
  eAscent = MAIN->EFONT->ascent;
  cAscent = MAIN->CFONT->ascent;
  
  isGB = (strcmp(MAIN->base_charset,"gb2312")==0);
  GB[2] = 0;

  while (*p!=0) {
    EnableChinese = MAIN->CC(p);
    if (EnableChinese>0) {
      /* Draw Chinese Font */
      if (isGB) {
        memcpy(GB,p,2);
        GB[0]-=0x80;
        GB[1]-=0x80;
        XDrawImageString16(MAIN->disp,MAIN->win,MAIN->CGC,x,y+cAscent,(XChar2b*)GB,1);
      } else
        XDrawImageString16(MAIN->disp,MAIN->win,MAIN->CGC,x,y+cAscent,(XChar2b*)p,1);
      x += cwidth;
      while (EnableChinese>0) { EnableChinese--; p++; }
    } else {
      /* Draw English Font */
      XDrawImageString(MAIN->disp,MAIN->win,MAIN->EGC,x,y+eAscent,p,1);
      x += ewidth;
      p++;
    }
  }
}

void a4cDrawTextInCenterBox(a4cMAIN *MAIN,a4cKeyBox *KeyBox)
{
  int x,y,w,sw;
  int ewidth,cwidth;
  int eAscent,cAscent;
  char * p = KeyBox->message;
  int EnableChinese;
  int isGB;
  char GB[3];

  XSetForeground(MAIN->disp,MAIN->EGC,MAIN->ENVCOLOR.englishcolor);
  XSetBackground(MAIN->disp,MAIN->EGC,MAIN->ENVCOLOR.background);
  XSetForeground(MAIN->disp,MAIN->CGC,MAIN->ENVCOLOR.foreground);
  XSetBackground(MAIN->disp,MAIN->CGC,MAIN->ENVCOLOR.background);

  ewidth = MAIN->EFONT->max_bounds.width;
  cwidth = MAIN->CFONT->max_bounds.width;
  eAscent = MAIN->EFONT->ascent;
  cAscent = MAIN->CFONT->ascent;

  isGB = (strcmp(MAIN->base_charset,"gb2312")==0);
  GB[2] = 0;

  /* Calculate String Width */
  sw = 0;
  while (*p!=0) {
    EnableChinese = MAIN->CC(p);

    if (EnableChinese>0) {
      sw += cwidth;
      while (EnableChinese>0) { EnableChinese--; p++; }
    } else {
      sw += ewidth;
      p++;
    }
  }

  p = KeyBox->message;
  x = (KeyBox->Text.x2+KeyBox->Text.x1-sw)/2;
  y = KeyBox->Text.y1;

  while (*p!=0) {
    EnableChinese = MAIN->CC(p);
    if (EnableChinese>0) {
      /* Draw Chinese Font */
      if (isGB) {
        memcpy(GB,p,2);
        GB[0]-=0x80;
        GB[1]-=0x80;
        XDrawImageString16(MAIN->disp,MAIN->win,MAIN->CGC,x,y+cAscent,(XChar2b*)GB,1);
      } else
        XDrawImageString16(MAIN->disp,MAIN->win,MAIN->CGC,x,y+cAscent,(XChar2b*)p,1);
      x += cwidth;
      while (EnableChinese>0) { EnableChinese--; p++; }
    } else {
      /* Draw English Font */
      XDrawImageString(MAIN->disp,MAIN->win,MAIN->EGC,x,y+eAscent,p,1);
      x += ewidth;
      p++;
    }
  }
}

void a4cDraw(a4cMAIN *MAIN)
{
  if (strcmp(MAIN->base_charset,MAIN->current_charset)!=0) {
    if (strcmp(MAIN->base_charset,"gb2312")==0) {
      
    } else
    if (strcmp(MAIN->base_charset,"big5")==0) {
      
    }
  }
  #define BORDER(item)\
    if (!COMPARE_FLAG(MAIN->item.style,A4C_STYLE_NOBORDER))\
      a4cDrawShadowBox(MAIN,&MAIN->item.Border,0)

  a4cDrawShadowBox(MAIN,&MAIN->Border,1);

  BORDER(Method);
  BORDER(KeyArea);
  BORDER(Charset);
  BORDER(FontSize);
  BORDER(ListArea);

  #undef  BORDER
  #define DRAWTEXT(item)\
    XSetForeground(\
      MAIN->disp,\
      MAIN->EGC,\
      MAIN->ENVCOLOR.background\
    );\
    XFillRectangle(\
      MAIN->disp,MAIN->win,MAIN->EGC,\
      MAIN->item.Text.x1,\
      MAIN->item.Text.y1,\
      MAIN->item.Text.x2-MAIN->item.Text.x1+1,\
      MAIN->item.Text.y2-MAIN->item.Text.y1+1\
    );\
    if COMPARE_FLAG(MAIN->item.style,A4C_STYLE_CENTER)\
      a4cDrawTextInCenterBox(MAIN,&MAIN->item);\
    else\
      a4cDrawTextInBox(MAIN,&MAIN->item)

  DRAWTEXT(Method);
  DRAWTEXT(KeyArea);
  DRAWTEXT(Charset);
  DRAWTEXT(FontSize);
  DRAWTEXT(ListArea);

  #undef  DRAWTEXT
}

void a4cMakeNoise(a4cMAIN *MAIN)
{
  if (MAIN->AUDIO_DEVICE==NULL||MAIN->WARNING_FILE==NULL)
    XBell(MAIN->disp,-50);
  else
    a4cPlayWAVE(MAIN->AUDIO_DEVICE,MAIN->WARNING_FILE);
}

int a4cWhere(a4cMAIN * MAIN,int x,int y)
{
  #define DETECT(item,id) if (isInKeyBox(&MAIN->item,x,y)) return id
  DETECT(Method,A4C_METHOD);
  DETECT(Charset,A4C_CHARSET);
  DETECT(FontSize,A4C_FONTSIZE);
  DETECT(KeyArea,A4C_KEYAREA);
  DETECT(ListArea,A4C_LISTAREA);
  #undef  DETECT
  return A4C_NOWHERE;
}

void WriteReply(a4cMAIN *MAIN,char *str,int len,int status)
{
  int slen = 0;
  MAIN->xcin_res.len    = len;
  MAIN->xcin_res.status = status;
  if (MAIN->Endian) {
    SwapEndian((char*)&MAIN->xcin_res.len);
    SwapEndian((char*)&MAIN->xcin_res.status);
  }
  memcpy(MAIN->xcin_res.key,str,len);
  slen = sizeof(XCIN_RES) - XCIN_KEY_LEN + len;
  XChangeProperty(
    MAIN->disp,
    MAIN->win,
    MAIN->XCIN,
    XA_STRING,
    8,
    PropModeReplace,
    (unsigned char *)&MAIN->xcin_res,
    slen
  );
  XSync(MAIN->disp,False);
}

void SendChineseKey(a4cMAIN *MAIN,unsigned char *s)
{
  WriteReply(MAIN,s,2,A4C_ACCEPT);
  MAIN->replied = 1;
}

void SendString(a4cMAIN *MAIN,unsigned char *s)
{
  if (MAIN->translating>1) {
    MAIN->translating=0;
    WriteReply(MAIN,s,strlen(s),A4C_ACCEPT);
  } else
  if (MAIN->current_filter!=0&&MAIN->FILTER_NOW!=NULL) {
    MAIN->translating = MAIN->FILTER_NOW->SIMPLE(&MAIN->FILTER_NOW->RECORD,MAIN->base_charset,s);
    if (MAIN->translating==1||
        MAIN->translating==2)   WriteReply(MAIN,s,strlen(s),A4C_ACCEPT);
                         else   SendZeroOut(MAIN);
  } else WriteReply(MAIN,s,strlen(s),A4C_ACCEPT);

  MAIN->replied = 1;
}

void SendChar(a4cMAIN *MAIN,char ch)
{
  WriteReply(MAIN,&ch,1,A4C_ACCEPT);
  MAIN->replied = 1;
}

void SendZeroOut(a4cMAIN *MAIN)
{
  WriteReply(MAIN,NULL,0,A4C_ACCEPT);
}

void SendReject(a4cMAIN *MAIN)
{
  WriteReply(MAIN,NULL,0,A4C_REJECT);
}

void SendFullSizeFont(a4cALL *ALL,char ch)
{
  char tmp[2];
  char * cc;
  tmp[0] = ch;
  tmp[1] = 0;
  cc = iomeWhichItem(&ALL->CONF.CONF,tmp);
  if (cc!=NULL) SendString(&ALL->MAIN,cc);
           else SendReject(&ALL->MAIN);
}

int a4cLoadModules(a4cMAIN *MAIN,iomeConfig *CONF,iomeConfig *CFG)
{
  int          n;
  iomeRegister REGF;
  iomeModule * MOD;
  iomeRegistry REG;
  a4cHANDLE    DLL;
  char *       DLLNAME;
  iomeSysInfo  INFO;

  if (CONF->size==0) return 0;
  INFO.system   = "x";
  INFO.platform = "a4c.ime";
  INFO.version  = "0.1a";
  INFO.charset  = MAIN->current_charset;
  MAIN->dl_handles = CONF->size;
  MAIN->SO_HANDLES = (a4cHANDLE*)malloc(sizeof(a4cHANDLE)*CONF->size);
  for (n=0;n<CONF->size;n++) 
  if (strcmp(CONF->items[n].key,"load")==0) {
    DLLNAME = CONF->items[n].value;
    MAIN->SO_HANDLES[n] = dlopen(DLLNAME,RTLD_NOW);
    DLL = MAIN->SO_HANDLES[n];
    REGF = dlsym(DLL,IME_REGISTER);
    if (REGF&&REGF(&INFO,&REG)) {
      /* See What Type of Module it is */
      if (strcmp(REG.class,"ime")==0) {
        /* Good, it's an input method, load it */
        /* Append a new method */
        MAIN->METHOD=(iomeModule**)realloc(MAIN->METHOD,sizeof(iomeModule*)*(MAIN->method_size+1));
        MOD=(iomeModule*)malloc(sizeof(iomeModule));
        MAIN->METHOD[MAIN->method_size]=MOD;
        MAIN->method_size++;
        /* Load Functions */
        memset(MOD,0,sizeof(iomeModule));

        #define LOADIT(fn,name)\
           MOD->fn = dlsym(DLL,name);\
           if (MOD->fn==NULL)\
             die("Can not load function %s in %s",name,DLLNAME);
        
        LOADIT(INIT,IME_INIT);
        LOADIT(PREPARE,IME_PREPARE);
        LOADIT(FEEDKEY,IME_FEEDKEY);
        LOADIT(LISTAREA,IME_LISTAREA);
        LOADIT(KEYAREA,IME_KEYAREA);
        LOADIT(FOCUS,IME_FOCUS);
        LOADIT(QUIT,IME_QUIT);

        #undef  LOADIT
        /* Initialize it */
        MOD->RECORD.MaxListArea = (A4C_KEYAREA_WIDTH+
                                   A4C_METHOD_WIDTH+
                                   A4C_CHARSET_WIDTH+
                                   A4C_FONTSIZE_WIDTH)*2;
        MOD->RECORD.MaxKeyArea  = A4C_KEYAREA_WIDTH * 2;
        MOD->INIT(CFG,&MOD->RECORD);

      } else
      if (strcmp(REG.class,"filter")==0) {
        /* Excellent, we have a filter */
        MAIN->FILTER = (iomeModule**)realloc(MAIN->FILTER,sizeof(iomeModule*)*(MAIN->filter_size+1));
        MOD = (iomeModule*)malloc(sizeof(iomeModule));
        MAIN->FILTER[MAIN->filter_size] = MOD;
        MAIN->filter_size++;
        /* Load Functions */
        #define LOADIT(fn,name)\
           MOD->fn = dlsym(DLL,name);\
           if (MOD->fn==NULL)\
             die("Can not load function %s in %s",name,DLLNAME);

        LOADIT(INIT,IME_INIT);
        LOADIT(SIMPLE,IME_SIMPLE);
        LOADIT(FEEDKEY,IME_FEEDKEY);
        LOADIT(QUIT,IME_QUIT);

        #undef  LOADIT
        /* Initialize it */
        MOD->RECORD.MaxListArea = (A4C_KEYAREA_WIDTH+
                                   A4C_METHOD_WIDTH+
                                   A4C_CHARSET_WIDTH+
                                   A4C_FONTSIZE_WIDTH)*2;
        MOD->RECORD.MaxKeyArea  = A4C_KEYAREA_WIDTH * 2;
        MOD->INIT(CFG,&MOD->RECORD);

      } else
        report("Cannot recognize %s modules type",CONF->items[n].value);
    }
  }
}

int a4cFreeModules(a4cMAIN *MAIN)
{
  int n;

  /* porque metodo cero esta ingles */
  for (n=1;n<MAIN->method_size;n++) {
    MAIN->METHOD[n]->QUIT(&MAIN->METHOD[n]->RECORD);
  }

  /* porque filtro cero esta nada */  
  for (n=1;n<MAIN->filter_size;n++) {
    MAIN->FILTER[n]->QUIT(&MAIN->FILTER[n]->RECORD);
  }

  while (MAIN->dl_handles>0) {
    MAIN->dl_handles--;
    dlclose(MAIN->SO_HANDLES[MAIN->dl_handles]);
  }
  return 0;
}

int UnifyCharset(a4cMAIN *MAIN,char *str)
{
  if (MAIN->translating>1) {
    if (strcmp(MAIN->base_charset,MAIN->FILTER_NOW->RECORD.Encoding)!=0) {
      if (strcmp(MAIN->base_charset,"gb2312")==0) {
        ConvertChinese(BIG5_GB2312,str);
      } else
      if (strcmp(MAIN->base_charset,"big5")==0) {
        ConvertChinese(GB2312_BIG5,str);
      }
    }
  } else
  if (strcmp(MAIN->base_charset,MAIN->current_charset)!=0) {
    if (strcmp(MAIN->base_charset,"gb2312")==0) {
      ConvertChinese(BIG5_GB2312,str);
    } else
    if (strcmp(MAIN->base_charset,"big5")==0) {
      ConvertChinese(GB2312_BIG5,str);
    }
  }
  return 0;
}
